package repositories.projectsrepository

import com.darkrockstudios.apps.hammer.common.components.storyeditor.metadata.Info
import com.darkrockstudios.apps.hammer.common.components.storyeditor.metadata.ProjectMetadata
import com.darkrockstudios.apps.hammer.common.data.ProjectDef
import com.darkrockstudios.apps.hammer.common.data.globalsettings.GlobalSettings
import com.darkrockstudios.apps.hammer.common.data.globalsettings.GlobalSettingsRepository
import com.darkrockstudios.apps.hammer.common.data.migrator.PROJECT_DATA_VERSION
import com.darkrockstudios.apps.hammer.common.data.projectmetadatarepository.ProjectMetadataDatasource
import com.darkrockstudios.apps.hammer.common.data.projectsrepository.ProjectsRepositoryOkio
import com.darkrockstudios.apps.hammer.common.dependencyinjection.createTomlSerializer
import com.darkrockstudios.apps.hammer.common.fileio.okio.toHPath
import com.darkrockstudios.apps.hammer.common.fileio.okio.toOkioPath
import com.darkrockstudios.apps.hammer.common.getDefaultRootDocumentDirectory
import io.mockk.Awaits
import io.mockk.coEvery
import io.mockk.coVerify
import io.mockk.every
import io.mockk.just
import io.mockk.mockk
import io.mockk.verify
import kotlinx.coroutines.flow.SharedFlow
import kotlinx.coroutines.test.advanceUntilIdle
import kotlinx.coroutines.test.runTest
import kotlinx.datetime.Instant
import net.peanuuutz.tomlkt.Toml
import okio.Path.Companion.toPath
import okio.fakefilesystem.FakeFileSystem
import org.junit.jupiter.api.BeforeEach
import org.junit.jupiter.api.Test
import org.koin.dsl.bind
import org.koin.dsl.module
import utils.BaseTest
import kotlin.test.assertTrue

class ProjectsRepositoryTest : BaseTest() {

	lateinit var ffs: FakeFileSystem
	lateinit var toml: Toml

	lateinit var globalSettingsRepository: GlobalSettingsRepository
	lateinit var settingsFlow: SharedFlow<GlobalSettings>

	private lateinit var metadataDatasource: ProjectMetadataDatasource
	lateinit var settings: GlobalSettings

	@BeforeEach
	override fun setup() {
		super.setup()

		ffs = FakeFileSystem()
		toml = createTomlSerializer()

		globalSettingsRepository = mockk()

		metadataDatasource = mockk(relaxed = true)
		every { metadataDatasource.loadMetadata(any()) } returns
			ProjectMetadata(
				info = Info(
					created = Instant.DISTANT_FUTURE,
					lastAccessed = Instant.DISTANT_FUTURE,
					dataVersion = PROJECT_DATA_VERSION,
				)
			)

		settings = GlobalSettings(
			projectsDirectory = (getDefaultRootDocumentDirectory().toPath() / "Projects").toHPath().path
		)
		every { globalSettingsRepository.globalSettings } returns settings

		settingsFlow = mockk()
		coEvery { settingsFlow.collect(any()) } just Awaits
		every { globalSettingsRepository.globalSettingsUpdates } returns settingsFlow

		val testModule = module {
			single { globalSettingsRepository } bind GlobalSettingsRepository::class
		}
		setupKoin(testModule)

		ffs.createDirectories(settings.projectsDirectory.toPath())
	}

	@Test
	fun `Initialize ProjectsRepository`() = runTest {
		val repo = ProjectsRepositoryOkio(
			fileSystem = ffs,
			globalSettingsRepository = globalSettingsRepository,
			projectsMetadataDatasource = metadataDatasource,
		)

		advanceUntilIdle()

		verify(exactly = 1) { globalSettingsRepository.globalSettingsUpdates }
		coVerify(exactly = 1) { settingsFlow.collect(any()) }
	}

	@Test
	fun `Create New Project`() = runTest {
		val repo = ProjectsRepositoryOkio(
			fileSystem = ffs,
			globalSettingsRepository = globalSettingsRepository,
			projectsMetadataDatasource = metadataDatasource,
		)

		val projectName = "Test Project"

		val result = repo.createProject(projectName)
		assertTrue { result.isSuccess }

		val projectPath = settings.projectsDirectory.toPath() / projectName
		ffs.exists(projectPath)

		val newDef = ProjectDef(projectName, projectPath.toHPath())
		val metadataPath = metadataDatasource.getMetadataPath(newDef)
		ffs.exists(metadataPath.toOkioPath())
	}
}